using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Serialization.Formatters.Binary;
using System.Xml;
using System.Xml.Serialization;
using WinFormsTicTacToe.GameClasses;


//====================================================
//       :
// https://www.interestprograms.ru
//     
//     
//====================================================

namespace WinFormsTicTacToe
{
    public partial class Form1 : Form
    {
        readonly GameCanvas gameCanvas = new(5);

        //     .
        readonly System.Windows.Forms.Timer timer = new();

        //      
        //    .     
        //    
        private Thread? threadReceive = null;
        private UdpClient? udp = null;

        //   
        private DataType GAMESTATE = DataType.Undefined;

        //      .
        private bool ISMOVE = false;



        public Form1()
        {
            InitializeComponent();

            //        .
            gameCanvas.Parent = this.panel1;
            gameCanvas.Width = panel1.Width;
            gameCanvas.Height = panel1.Height;

            //      
            gameCanvas.EventPlayerMove += GameCanvas_EventPlayerMove;
            //     
            gameCanvas.EventGameVictory += GameCanvas_EventVictory;
            //     
            gameCanvas.EventGameDraw += GameCanvas_EventDraw;

            //   
            timer.Tick += Timer_Tick;
            timer.Stop();

            
        }


        private void Form1_Shown(object sender, EventArgs e)
        {
            //       
            //  Load    .

            //  TopMost
            // https://stackoverflow.com/questions/16862057/this-topmost-true-not-working
            // https://support.microsoft.com/en-us/topic/topmost-windows-are-not-always-in-the-topmost-position-in-windows-7-or-in-windows-server-2008-r2-24fb5117-0e97-da0a-874f-ef95122ef7cb
            this.TopMost = true;
        }


        #region     

        private void Timer_Tick(object? sender, EventArgs e)
        {
            //     .
            SendInitGame();

            //     .
            Console.Beep(2000, 5);
        }


        private void BtnStartGame_Click(object sender, EventArgs e)
        {
            //    
            //        
            //  .
            btnStartGame.Enabled = false;

            //    .
            displayBoard.Text = "  ...";

            //    
            EnabledNetworkControl(false);

            //  UDP    
            //    .
            if (udp == null)
            {
                //  -   .
                StartReceive();
            }

            //    
            //     .
            int[] timeIntervals = new[] { 500, 3000 };
            timer.Interval = timeIntervals[new Random().Next(timeIntervals.Length)];

            //     
            timer.Start();
        }


        private void Form1_Closing(object sender, FormClosingEventArgs e)
        {
            //   ,
            //   .
            if (udp != null)
            {
                udp.Close();
                udp = null;
            }

            //  ,   
            //       .
            if (threadReceive != null) threadReceive.Join();

            //        .
            //      
            //       .
            DataChange change = new()
            {
                gameState = DataType.AppClose
            };
            Send(change);
        }


        //        .
        private void PanelFigure_Paint(object sender, PaintEventArgs e)
        {
            if (sender is Panel panel)
            {
                Brush rcBrush = Constants.rcBrush;
                Brush figureBrush = Constants.cellsBrush;

                //     .
                if (gameCanvas.ActiveState == false)
                {
                    rcBrush = Constants.rcBrushDeactive;
                    figureBrush = Constants.cellsBrushDeactive;
                }

                //    .
                Graphics g = e.Graphics;
                g.FillRectangle(rcBrush, 0, 0, panel.Width, panel.Height);

                //     .
                if (GAMESTATE != DataType.Undefined)
                {
                    Figures f = new()
                    {
                        type = 2
                    };
                    if (gameCanvas._cross == true) f.type = 1;

                    Point origin = new(panel.Width / 2, panel.Width / 2);
                    f.Draw(g, origin, panel.Width, figureBrush);
                }
            }
        }

        //     .
        //      .
        private void EnabledNetworkControl(bool enabled)
        {
            tbIPRemote.Enabled = enabled;
            tbPortRemote.Enabled = enabled;
            tbPortHost.Enabled = enabled;
        }

        #endregion


        #region    

        //   
        private void GameCanvas_EventDraw(object? sender, EventArgs e)
        {
            //    - .
            StateEndGame(DataType.Draw, ",  !");

            //        ,
            //   ,  .
            DataChange change = new()
            {
                gameState = DataType.Draw
            };
            Send(change);
        }


        //   
        private void GameCanvas_EventVictory(object? sender, EventArgs e)
        {
            //      .
            StateEndGame(DataType.Victory, " !");

            //  -     
            //  .
            DataChange change = new()
            {
                gameState = DataType.Victory
            };
            Send(change);
        }


        //    
        private void GameCanvas_EventPlayerMove(object? sender, PlayerMoveEventArgs e)
        {
            //      .
            if (GAMESTATE == DataType.Game)
            {
                if (ISMOVE == true)
                {
                    //      .
                    int type = 2;
                    if (gameCanvas._cross == true) type = 1;

                    //      .
                    gameCanvas.DoDraw(e.X, e.Y, type);

                    //        .
                    ISMOVE = false;

                    //       .
                    displayBoard.Text = "  !";

                    //     
                    //      ,
                    //        .
                    DataChange change = new()
                    {
                        gameState = DataType.PlayerMove,
                        X = e.X,
                        Y = e.Y,
                        type = (byte)type
                    };

                    Send(change);
                }
                else
                {
                    MessageBox.Show("  ");
                }
            }
            else
            {
                MessageBox.Show(" !");
            }

        }

        #endregion


        #region  

        private void Send(DataChange change)
        {
            //***************

            // https://docs.microsoft.com/ru-ru/dotnet/standard/serialization/binaryformatter-security-guide

            //  1
            MemoryStream memoryStream = new();
            BinaryFormatter bf = new();
            bf.Serialize(memoryStream, change);


            //  2
            /*
            //https://stackoverflow.com/questions/7442164/c-sharp-and-net-how-to-serialize-a-structure-into-a-byte-array-using-binary
            XmlSerializer serializer = new(typeof(DataChange));
            MemoryStream memoryStream = new();
            XmlWriter xmlWriter = XmlWriter.Create(memoryStream);
            serializer.Serialize(xmlWriter, change);
            */

            byte[] sent = memoryStream.ToArray();

            //     UdpClient  
            //   
            UdpClient udp = new();

            //      
            //   ,  ip  
            //     .
            //   ,   .
            IPEndPoint ep = new(IPAddress.Parse(tbIPRemote.Text), int.Parse(tbPortRemote.Text));
            udp.Connect(ep);
            udp.Send(sent, sent.Length);

            //  UDP 
            udp.Close();
        }


        private void StartReceive()
        {
            //      ,
            //      .
            ThreadStart tstart = new(Receive);
            threadReceive = new Thread(tstart);
            threadReceive.Start();
        }


        //     
        //     
        private void Receive()
        {
            try
            {
                //       .
                //         .
                IPEndPoint ep = new(IPAddress.Any, int.Parse(tbPortHost.Text));
                udp = new UdpClient(ep);
                while(udp != null)
                {
                    IPEndPoint? remote = null;
                    byte[] recv = udp.Receive(ref remote);

                    //  1
                    MemoryStream memoryStream = new(recv);
                    BinaryFormatter bf = new();
                    DataChange? change = (DataChange)bf.Deserialize(memoryStream);

                    //  2
                    /* XmlSerializer serializer = new(typeof(DataChange));
                     MemoryStream memoryStream = new(recv);
                     XmlReader xmlReader = XmlReader.Create(memoryStream);
                     DataChange? change = (DataChange?)serializer.Deserialize(xmlReader);*/

                    //      .
                    if (change != null) ParseReceive(change);
                }
            }
            //      UDP .
            // https://docs.microsoft.com/ru-ru/dotnet/csharp/language-reference/keywords/try-catch
            catch (SocketException e) when (e.ErrorCode == (int)SocketError.Interrupted)
            {
                
            }
            catch { }
            finally
            {
                //       
                //   
                if (udp != null)
                {
                    udp.Close();
                    udp = null;
                }
            }
        }


        //          .
        //       
        delegate void ReceiveCallback(DataChange change);


        //   
        void ParseReceive(DataChange change)
        {
            if (this.displayBoard.InvokeRequired)
            {
                ReceiveCallback dt = new(ParseReceive);
                this.Invoke(dt, new object[] { change });
            }
            else
            {
                switch (change.gameState)
                {
                    //     (  )
                    case DataType.Start:
                        {
                            //     
                            //    
                            //  .
                            timer.Stop();

                            //    
                            //     
                            //   .
                            //    .
                            if (GAMESTATE != DataType.Game)
                            {
                                StateGame(false, " !   !");

                                //  -  .
                                change.gameState = DataType.Game;
                                Send(change);
                            }
                            //else MessageBox.Show("GameState.Start");
                        }
                        break;

                    //  -  .
                    case DataType.Game:

                        //       
                        //    .
                        timer.Stop();

                        //    
                        //     
                        //   .
                        //    .
                        if (GAMESTATE != DataType.Game) StateGame(true, " !");
                        //else MessageBox.Show("GameState.Game");


                        break;

                    //       .
                    case DataType.PlayerMove:
                        //       
                        //    .
                        gameCanvas.DoDraw(change.X, change.Y, change.type);

                        //     -,
                        //     .
                        ISMOVE = true;

                        //       .
                        displayBoard.Text = " !";

                        break;

                    //         .
                    case DataType.Victory:
                    
                        StateEndGame(DataType.Victory, " ,  ...");

                        break;

                    //        - .
                    case DataType.Draw:
                       
                        StateEndGame(DataType.Draw, ",  !");

                        break;
                    case DataType.AppClose:

                        StateEndGame(DataType.AppClose, "    !");

                        break;
                }
            }
        }


        //    .
        private void SendInitGame()
        {
            //      .
            DataChange datachange = new()
            {
                //   - 
                gameState = DataType.Start
            };

            //      .
            Send(datachange);
        }

        #endregion


        #region   

        /// <summary>
        ///   
        /// </summary>
        private void StateGame(bool ismove, string hint)
        {
            //    :
            //     ,
            //      .
            btnStartGame.Enabled = false;

            //   
            GAMESTATE = DataType.Game;

            //       
            //   .
            gameCanvas.Clear();

            //      (true)  (false).
            //     .
            gameCanvas._cross = ismove;

            //   ,      .
            //   (true)    (false).
            ISMOVE = ismove;

            //      .
            displayBoard.Text = hint;

            //   
            panelFigure.Invalidate();
        }

        //      .
        private void StateEndGame(DataType state, string hint)
        {
            //     
            GAMESTATE = state;

            //      
            gameCanvas.ActiveState = false;

            //     
            btnStartGame.Enabled = true;

            //     .
            displayBoard.Text = hint;

            //   
            panelFigure.Invalidate();
        }

        #endregion


        #region  

        private void LinkLabel1_Clicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            string link = ((LinkLabel)sender).Text;
            Process.Start(new ProcessStartInfo(link) { UseShellExecute = true });
        }

        #endregion
    }
}